from pliers.support.due import due, Url

due.cite(
    Url("https://github.com/akisato-/pySaliencyMap"),
    description="pySaliencyMap module -- Image saliency estimation",
    path='pySaliencyMap',
    cite_module=True,
    tags=["reference-implementation"]
)